"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RpcApiCaller = void 0;
var tslib_1 = require("tslib");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var constants_1 = require("./constants");
var BufferSubject_1 = require("../../../util/BufferSubject");
var error_1 = require("./error");
var RpcApiCaller = (function () {
    function RpcApiCaller(_a) {
        var api = _a.api, _b = _a.preCallBufferSize, preCallBufferSize = _b === void 0 ? 10 : _b, _c = _a.maxActiveCalls, maxActiveCalls = _c === void 0 ? 50 : _c;
        this._calls = 0;
        this.api = api;
        this.preCallBufferSize = preCallBufferSize;
        this.maxActiveCalls = maxActiveCalls;
    }
    Object.defineProperty(RpcApiCaller.prototype, "activeCalls", {
        get: function () {
            return this._calls;
        },
        enumerable: false,
        configurable: true
    });
    RpcApiCaller.prototype.exists = function (name) {
        return this.api.hasOwnProperty(name);
    };
    RpcApiCaller.prototype.get = function (name) {
        if (!this.exists(name))
            throw new error_1.RpcError(constants_1.RpcServerError.NoMethodSpecified);
        return this.api[name];
    };
    RpcApiCaller.prototype.call = function (name, request, ctx) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var method;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                if (this._calls >= this.maxActiveCalls)
                    throw new error_1.RpcError(constants_1.RpcServerError.TooManyActiveCalls);
                method = this.get(name);
                if (method.validate) {
                    try {
                        method.validate(request);
                    }
                    catch (error) {
                        throw new error_1.RpcValidationError(error);
                    }
                }
                this._calls++;
                return [2, (method.onPreCall ? method.onPreCall(ctx, request) : Promise.resolve())
                        .then(function () { return !method.isStreaming
                        ? method.call(ctx, request)
                        : rxjs_1.firstValueFrom(method.call$(ctx, rxjs_1.of(request))); })
                        .finally(function () {
                        _this._calls--;
                    })];
            });
        });
    };
    RpcApiCaller.prototype.createCall = function (name, ctx) {
        var _this = this;
        var req$ = new rxjs_1.Subject();
        var reqUnsubscribe$ = new rxjs_1.Subject();
        try {
            if (this._calls >= this.maxActiveCalls)
                throw new error_1.RpcError(constants_1.RpcServerError.TooManyActiveCalls);
            var method = this.get(name);
            if (!method.isStreaming) {
                var response$ = rxjs_1.from((function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                    var request, response;
                    return tslib_1.__generator(this, function (_a) {
                        switch (_a.label) {
                            case 0: return [4, rxjs_1.firstValueFrom(req$.pipe(operators_1.first()))];
                            case 1:
                                request = _a.sent();
                                return [4, this.call(name, request, ctx)];
                            case 2:
                                response = _a.sent();
                                return [2, response];
                        }
                    });
                }); })());
                var res$_1 = new rxjs_1.Subject();
                response$.subscribe(res$_1);
                return { req$: req$, reqUnsubscribe$: reqUnsubscribe$, res$: res$_1 };
            }
            var methodStreaming_1 = method;
            var requestValidated$ = req$
                .pipe(operators_1.map(function (request) {
                try {
                    if (methodStreaming_1.validate)
                        methodStreaming_1.validate(request);
                    return request;
                }
                catch (error) {
                    throw new error_1.RpcValidationError(error);
                }
            }));
            var bufferSize = methodStreaming_1.preCallBufferSize || this.preCallBufferSize;
            var requestBuffered$_1 = new BufferSubject_1.BufferSubject(bufferSize);
            requestBuffered$_1.subscribe({ error: function () { } });
            var error$_1 = new rxjs_1.Subject();
            requestBuffered$_1.subscribe({
                error: function (error) { error$_1.error(error); },
            });
            requestValidated$.subscribe(requestBuffered$_1);
            var result$_1 = requestBuffered$_1
                .pipe(operators_1.take(1), operators_1.switchMap(function (request) {
                return methodStreaming_1.onPreCall
                    ? rxjs_1.from(methodStreaming_1.onPreCall(ctx, request))
                    : rxjs_1.from([0]);
            }), operators_1.switchMap(function () {
                Promise.resolve().then(function () {
                    requestBuffered$_1.flush();
                });
                return methodStreaming_1.call$(ctx, requestBuffered$_1)
                    .pipe(operators_1.finalize(function () {
                    error$_1.complete();
                }));
            }), operators_1.share(), operators_1.mergeWith(error$_1));
            this._calls++;
            var resultWithActiveCallTracking$_1 = result$_1.pipe(operators_1.finalize(function () {
                error$_1.complete();
                _this._calls--;
            }));
            var res$ = new rxjs_1.Observable(function (observer) {
                var _a;
                var subscription = resultWithActiveCallTracking$_1.subscribe(observer);
                var timeout = (_a = methodStreaming_1.timeout) !== null && _a !== void 0 ? _a : 15000;
                var timeoutSubscription = requestBuffered$_1
                    .pipe(operators_1.catchError(function () { return rxjs_1.EMPTY; }), operators_1.mergeWith(result$_1.pipe(operators_1.catchError(function () { return rxjs_1.EMPTY; }))), operators_1.debounce(function () { return rxjs_1.interval(timeout); }))
                    .subscribe(function () {
                    var error = new error_1.RpcError(constants_1.RpcServerError.Timeout);
                    error$_1.error(error);
                });
                return function () {
                    timeoutSubscription.unsubscribe();
                    subscription.unsubscribe();
                };
            });
            return { req$: req$, reqUnsubscribe$: reqUnsubscribe$, res$: res$ };
        }
        catch (error) {
            req$.error(error);
            var res$ = new rxjs_1.Subject();
            res$.error(error);
            return { req$: req$, reqUnsubscribe$: reqUnsubscribe$, res$: res$ };
        }
    };
    RpcApiCaller.prototype.call$ = function (name, request$, ctx) {
        var call = this.createCall(name, ctx);
        request$.subscribe(call.req$);
        return call.res$;
    };
    return RpcApiCaller;
}());
exports.RpcApiCaller = RpcApiCaller;
